#include "config.h"
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#define XK_LATIN1
#include <X11/keysymdef.h>
#ifdef HAVE_X11_EXTENSIONS_XINERAMA_H
#include <X11/extensions/Xinerama.h>
#endif
#include "defs.h"
#include "ltdl.h"
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/time.h>
#ifdef TIME_WITH_SYS_TIME
#include <time.h>
#endif
#include <sys/types.h>
#include <sys/wait.h>
#include <pwd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/resource.h>
#define X_MAIN
#include "pixmap.h"
#include "globals.h"
#include "liblist.h"
#include "atoms.h"
#include "xhelp.h"
#include "tasklist.h"
#include "font.h"
#include <getopt.h>
#include "menu.h"
#include "menuhandler.h"
#include "tooltip.h"
#include "options.h"

/* pixmap data */
#include "icon.xpm"
#include "start.xpm"

#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlen
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

#define DEFAULT_BUTTON_WIDTH 150
#define ICON_WIDTH 16
#define ICON_HEIGHT 16

LPFONT pFont;
static LPLIBLIST pLibList;
static int textY;
static LPTASK taskHead;
static LPTASK currentFocus;
static int numTasks;
static char fontName[256];
static LPCONFIGLIST pConfigFile = NULL;
static LPMENU pMenu = NULL;
/* the x position of the start of the module area */
static int xModuleStart;
/* the x position of the end of the module area */
static int xModuleEnd;
/* the width of each task button */
static int buttonWidth = DEFAULT_BUTTON_WIDTH;
/* the beginning width where the first button may be drawn */
static int startButtonAreaWidth = 0;
/* tooltip flag */
static LPMODULE pTooltipModule = NULL;
static LPTOOLTIP pTooltip = NULL;
static PIXMAP startMenuIcon;

static int WND_HEIGHT=26;
#define SUBWND_HEIGHT (WND_HEIGHT - 2*(OFFSET+1))
#define OFFSET 2
#define TEXT_OFFSET 5
#define DEFAULT_FONT_NAME "-adobe-helvetica-medium-r-normal-*-12-*"
#define DEFAULT_FONT_NAME_XFT "times,serif-12"

static void LoadConfigFile();

static void AlarmHandler(int sig)
{
  Window root_ret, child_ret;
  int x,y,win_x,win_y;
  unsigned int mask;
  const char *pTxt;
  
  if (!pTooltip && pTooltipModule && pTooltipModule->GetTooltip &&
      (pTxt = pTooltipModule->GetTooltip(pTooltipModule)) && *pTxt)
  {
    XQueryPointer(pDisp,root,&root_ret,&child_ret,&x,&y,&win_x,&win_y,&mask);
    pTooltip = CreateTooltip(pTxt,x,y,PALETTE_TOOLTIP,
			     onTop ? TOOLTIP_SHOW_BOTTOM_LEFT : TOOLTIP_SHOW_TOP_LEFT,
			     pFont);
  }
}

static void ChildHandler(int sig)
{
  int status;
  /* get the status of the child that just died */
  wait(&status);
#ifdef DEBUG
  printf("Child quit, status is %i\n",status);
#endif
}

static void ReinitHandler(int sig)
{
  /* now load the file */
  LoadConfigFile();  
  /* reload the menu */
  if (pMenu)
  {
    DestroyMenu(pMenu);
    pMenu = NULL;
  }
}

static void Handler(int sig)
{
#ifdef DEBUG
  printf("Received signal, shutting down\n");
#endif
  
  if (pConfigFile)
  {
    ListDestroy((LPLIST)pConfigFile);
  }

  if (pMenu)
  {
    DestroyMenu(pMenu);
  }

  if (pLibList)
  {
    LibListDestroy(pLibList);	/* destroy the library list */
  }

  lt_dlexit();

  /* remove the tasks */
  RemoveAllTasks(&taskHead);

  if (pFont)
  {
    pFont->Release(pFont);
  }
  
  if (mainWindow)
  {
    XDestroyWindow(pDisp,mainWindow);
  }
  
  FreePixmap(&defaultPixmap);
  FreePixmap(&startMenuIcon);

  if (gGc)
  {
    XFreeGC(pDisp,gGc);
  }

  if (pDisp)			/* destroy the display */
  {
    XCloseDisplay(pDisp);
  }

  DestroyPixmap();
  /* and exit */
  exit(0);
}

static int IOErrorHandler(Display *pDisplay)
{
  Handler(SIGINT);
  return 0;
}

static int ErrorHandler(Display *pDisplay,XErrorEvent *event)
{
  /*char err[256];
  XGetErrorText(pDisplay,event->error_code,err,sizeof(err));
  printf("X error: %s\n",err); */
  return 0;
}

static void InstallSignals()
{
  signal(SIGINT,Handler);
  signal(SIGQUIT,Handler);
  signal(SIGTERM,Handler);
  signal(SIGHUP,ReinitHandler);
  signal(SIGCHLD,ChildHandler);
}

static void PrintDisplayError()
{
  const char *lpszDisp = getenv("DISPLAY");
  if (lpszDisp)
  {
    fprintf(stderr,"Unable to open: %s\n",lpszDisp);
  }
  else
  {
    fprintf(stderr,"You don't have $DISPLAY set to anything\n");
  }
}

static void PrintUsage(const char *progName,FILE *out)
{
  fprintf(out,"Usage: %s [OPTION]...\nDisplays a Win95 style taskbar in X\n\n"
	  "  -t, --top                    puts the toolbar on the top of the screen\n"
#ifdef HAVE_XFT
	  "  -x, --xft                    enable the use of Xft antialiased fonts\n"
#endif
	  "  -d, --daemon                 run as a daemon\n"
	  "  -f, --font NAME              sets the font used\n"
	  "  -m, --enable-menu            enables the start menu\n"
	  "  -a, --fancy-menu             enables Office 2000 style fancy menus\n"
	  "  -s, --disable-tooltips       disables tooltips in the module window\n"
	  "  -b, --disable-module-border  does not draw a rectangular module border\n"
	  "  -i, --disable-menu-icons     do not use icons in the start menu\n"
	  "  -h, --help                   displays this help\n",progName);
}

LPCONFIGLIST parse(const char *pFile);

static void LoadConfigFile()
{
  char fName[256];
  const char *pHome;
  int c;
  
  pHome = getenv("HOME");
  if (pHome)
  {
    c = strlen(pHome);
    if (pHome[c-1] == '/')
    {
      sprintf(fName,"%s%s",pHome,".taskbar");
    }
    else
    {
      sprintf(fName,"%s/%s",pHome,".taskbar");
    }
  }
  else
  {
    strcpy(fName,".taskbar");
  }
  /* now load the file */
  if (pConfigFile)
  {
    ListDestroy((LPLIST)pConfigFile);
  }
  pConfigFile = parse(fName);

}
static int LoadSettings(int argc,char **argv)
{
  static struct option long_opts [] = {
#ifdef HAVE_XFT
    { "xft"   , no_argument,       0, 'x' },
#endif
    { "daemon", no_argument,       0, 'd' },
    { "font"  , required_argument, 0, 'f' },
    { "help"  , no_argument,       0, 'h' },
    { "enable-menu"  , no_argument, 0, 'm' },
    { "disable-tooltips", no_argument, 0, 's'},
    { "top", no_argument, 0 , 't'},
    { "fancy-menu", no_argument,0,'a'},
    { "disable-module-border",no_argument, 0, 'b'},
    { "disable-menu-icons", no_argument, 0, 'i'},
    { 0,0,0,0 }
  };
  
#ifdef HAVE_XFT
#define ACCEPTED_OPTS "xdhmtasbif:"
#else
#define ACCEPTED_OPTS "dhmtasbif:"
#endif
  int c;
  int option_index;
  
  /* load the defaults */
  useXft = 0;
  useDaemon = 0;
  useMenu = 0;
  useMenuIcons = 1;
  useTooltips = 1;
  useModuleBorder = 1;
  onTop = 0;
  menuStyle = MS_NORMAL;

  fontName[0] = 0;

  while ((c=getopt_long(argc,argv,ACCEPTED_OPTS,long_opts,&option_index)) != -1)
  {
    switch (c)
    {
#ifdef HAVE_XFT
    case 'x':
      useXft = 1;
      break;
#endif
    case 'd':
      useDaemon = 1;
      break;
    case 'i':
      useMenuIcons = 0;
      break;
    case 'b':
      useModuleBorder = 0;
      break;
    case 'a':
      menuStyle = MS_FANCY;
      break;
    case 'm':
      useMenu = 1;
      break;
    case 't':
      onTop = 1;
      break;
    case 's':
      useTooltips = 0;
      break;
    case 'f':
      strcpy(fontName,optarg);
      break;
    case 'h':
    default:
      PrintUsage(argv[0],stdout);
      return 0;
    }
  }
  
  if (!(*fontName))		/* load the default font */
  {
#ifdef HAVE_XFT
    if (useXft)
    {
      strcpy(fontName,DEFAULT_FONT_NAME_XFT);
    }
    else
#endif
    {
      strcpy(fontName,DEFAULT_FONT_NAME);
    }
  }

  /* now parse the config file */
  LoadConfigFile();
  return 1;
}

static void LoadPalette()
{
  static unsigned short desired[][3] = { 
    /*{ 0xbefb,0xbaea,0xbefb }, */
    { 0xD600,0xD600,0xD600 },
    { 0x8617,0x8207,0x8617 },    
    { 0,0,0 },
    { 0xFFFF,0xFFFF,0xFFFF },
    { 0x22FF,0x37FF, 0x9FFF},
    { 0xFFFF,0xFFFF, 0xE0FF},
    { 0xEAEA,0xEAEA, 0xEAEA}
  };

  unsigned int i;
  for (i=0;i<sizeof(desired)/sizeof(desired[0]);++i)
  {
    palette[i] = GetColor(desired[i][0],desired[i][1],desired[i][2]);
  }
}

static void DrawButton(Drawable drawable,int x,int y,int width,int height,int up,int drawBackground)
{
  if (drawBackground)
  {
    XClearArea(pDisp,drawable,x,y,width,height,False);
  }
  /* draw the Win95 type border */
  /* white lines first */
  if (up)
  {
    XSetForeground(pDisp,gGc,PALETTE_WHITE);
    XDrawLine(pDisp,drawable,gGc,x,y,x,height-1);
    XDrawLine(pDisp,drawable,gGc,x,y,x+width-1,y);
    /* now gray lines */
    XSetForeground(pDisp,gGc,PALETTE_DARK_GRAY);
    XDrawLine(pDisp,drawable,gGc,x+1,height-1,x+width-1,height-1);
    XDrawLine(pDisp,drawable,gGc,x+width-1,y+1,x+width-1,height-1);
    /* now black lines */
    XSetForeground(pDisp,gGc,PALETTE_BLACK);
    XDrawLine(pDisp,drawable,gGc,x,height,x+width,height);
    XDrawLine(pDisp,drawable,gGc,x+width,y,x+width,height);
  }
  else
  {
    XSetForeground(pDisp,gGc,PALETTE_BLACK);
    XDrawLine(pDisp,drawable,gGc,x,y,x,height-1);
    XDrawLine(pDisp,drawable,gGc,x,y,x+width-1,y);
    /* now gray lines */
    XSetForeground(pDisp,gGc,PALETTE_DARK_GRAY);
    XDrawLine(pDisp,drawable,gGc,x+1,y+1,x+width-1,y+1);
    XDrawLine(pDisp,drawable,gGc,x+1,y+1,x+1,height-1);
    /* now white lines */
    XSetForeground(pDisp,gGc,PALETTE_WHITE);
    XDrawLine(pDisp,drawable,gGc,x,height,x+width,height);
    XDrawLine(pDisp,drawable,gGc,x+width,y,x+width,height);
  }
}

static void DrawModuleRectangle()
{
  /* now white lines */
  XSetForeground(pDisp,gGc,PALETTE_WHITE);
  XDrawLine(pDisp,mainWindow,gGc,xModuleStart,WND_HEIGHT-2*OFFSET+1,xModuleEnd,WND_HEIGHT-2*OFFSET+1);
  XDrawLine(pDisp,mainWindow,gGc,xModuleEnd,OFFSET,xModuleEnd,WND_HEIGHT-2*OFFSET+1);
  /* now gray lines */
  XSetForeground(pDisp,gGc,PALETTE_DARK_GRAY);
  XDrawLine(pDisp,mainWindow,gGc,xModuleStart,OFFSET,xModuleStart,WND_HEIGHT-2*OFFSET+1);
  XDrawLine(pDisp,mainWindow,gGc,xModuleStart,OFFSET,xModuleEnd-1,OFFSET);
}

static void DrawTask(LPTASK pTask,int x,int width,int bErase)
{
  int textSpace;
  int numChars;
  int textOffset;
#ifdef DEBUG
  printf("Drawing task 0x%08X: %s:%i\n",(unsigned int)pTask,pTask->name,pTask->nameLen);
#endif
  /* fill it with gray */
  if (bErase)
  {
    XClearArea(pDisp,mainWindow,x,onTop ? 0 : 1,width,WND_HEIGHT-1,False);
  }
  /* now draw the white checks if focused */
  if (pTask->dwFlags & TF_FOCUS)
  {
    XSetForeground(pDisp,gGc,PALETTE_LIGHTEST_GRAY);
    XFillRectangle(pDisp,mainWindow,gGc,x+OFFSET,OFFSET,width - 2*OFFSET,WND_HEIGHT - 2*OFFSET);
  }
  /* now draw the button */
  DrawButton(mainWindow,x+OFFSET,OFFSET,
	     width - 2 * OFFSET,WND_HEIGHT - 2*OFFSET + 1,
	     (pTask->dwFlags & TF_FOCUS) ? 0 : 1,0);
  
  if (pTask->dwFlags & TF_ICONIFIED)
  {
    pFont->SetColor(pFont,PALETTE_DARK_GRAY);
  }
  else
  {
    pFont->SetColor(pFont,PALETTE_BLACK);
  }
  /* limit myself to the width */
  numChars = pTask->nameLen;
  textSpace = width - 5 - OFFSET;
  if (pTask->icon.icon.icon != None)
  {
    textSpace -= ICON_WIDTH + OFFSET;
    textOffset = x + TEXT_OFFSET + ICON_WIDTH + OFFSET;
  }
  else
  {
    textOffset = x + TEXT_OFFSET;
  }
  
  while ((numChars > 0) && (pFont->GetWidth(pFont,pTask->name,numChars) >= textSpace))
  {
    --numChars;
  }

  if (numChars > 0)
  {
    pFont->Render(pFont,mainWindow,textOffset,textY,pTask->name,numChars);
  } 

  if (pTask->icon.icon.icon != None)
  {
    RenderPixmap(pDisp,mainWindow,gGc,
		 x+TEXT_OFFSET,(WND_HEIGHT-ICON_HEIGHT)/2,
		 &pTask->icon.icon);
  }
}

static void Redraw(LPTASK pTask)
{
  /* determine the X coord of it */
  int pos = 0;
  LPTASK iter;
  for (iter=taskHead;iter && (iter != pTask);iter=iter->next)
  {
    ++pos;
  }
#ifdef DEBUG
  printf("Redrawing position %i\n",pos);
#endif
  pos = startButtonAreaWidth + buttonWidth * pos;
  DrawTask(pTask,pos,buttonWidth,True);
}

static void DrawStartButton()
{
  int up = pMenu ? !pMenu->shown : 1;
  int shift = up ? 0 : 1;
  /* draw the button */
  DrawButton(mainWindow,OFFSET,OFFSET,
	     startButtonAreaWidth - 2 * OFFSET,WND_HEIGHT - 2*OFFSET+1,
	     up,True);
  /* draw the icon */
  RenderPixmap(pDisp,mainWindow,gGc,
	       2*OFFSET + shift,(WND_HEIGHT-ICON_HEIGHT)/2 + shift,
	       &startMenuIcon);
  
  /* and now the text */
  pFont->SetColor(pFont,PALETTE_BLACK);
  pFont->Render(pFont,mainWindow,2*OFFSET + startMenuIcon.width + shift,textY + shift,"Start",5);
}

static void Draw()
{
  LPTASK iter;
  int x = startButtonAreaWidth;
  int spaceRequired = xModuleStart;
  
  /* Erase everything up to the spaceRequired */
  XClearArea(pDisp,mainWindow,0,0,spaceRequired,WND_HEIGHT,False);
  /* draw top or bottom line for 3d effect */
  XSetForeground(pDisp,gGc,PALETTE_WHITE);
  XDrawLine(pDisp,mainWindow,gGc,0,onTop ? WND_HEIGHT - 1 : 0,
	    gScreenWidth,onTop ? WND_HEIGHT - 1 : 0);
  /* draw the start menu button */
  if (useMenu)
  {
    DrawStartButton();
  }
  /* draw the thing around the modules */
  if (useModuleBorder)
  {
    DrawModuleRectangle();
  }
  /* iterate through all the items */
  for (iter = taskHead;iter;iter = iter->next)
  {
#ifdef DEBUG
    printf("Drawing task at %i\n",x);
#endif
    DrawTask(iter,x,buttonWidth,False);
    x += buttonWidth;
  }
}

static void RecalcGeometry()
{
  LPTASK iter;
  int spaceRequired;

  for (numTasks=0,iter = taskHead;iter;iter = iter->next)
  {
    ++numTasks;
  }
  /* determine the width of each button */
  spaceRequired = xModuleStart - startButtonAreaWidth;
  if ((numTasks * DEFAULT_BUTTON_WIDTH) < spaceRequired)
  {
    buttonWidth = DEFAULT_BUTTON_WIDTH;
  }
  else
  {
    buttonWidth = spaceRequired / numTasks;
  }
}


static void BuildClientList(int forceRedraw)
{
  Window *pWin,focus;
  LPTASK iter;
  int rev;
  unsigned int flags;
  int drawRequired = forceRedraw;
  FINDTASK ft;
  
  XGetInputFocus(pDisp,&focus,&rev);
  
  pWin = (Window*)GetPropertyData(root,atom__NET_CLIENT_LIST,XA_WINDOW,&numTasks);
  if (pWin)
  {
#ifdef DEBUG
    printf("Enumerating windows\n");
#endif
    UnmarkAllTasks(taskHead);	/* unmark them all */
    /* now go through and add them all */
    for (rev=0;rev < numTasks;++rev)
    {
      /* try to find it first */
      flags = TF_VALID | ((pWin[rev] == focus) ? TF_FOCUS : 0);
      if (IsIconified(pWin[rev]))
      {
	flags |= TF_ICONIFIED;
      }

      /* now attempt to find it in our list */
      ft.dwFlags = FTF_WINDOW;
      ft.wnd     = pWin[rev];
      iter = FindTask(taskHead,&ft);
      if (!iter)
      {
	if ((iter=AddTask(&taskHead,pWin[rev],flags)))
	{
	  if (flags & TF_FOCUS)
	  {
	    currentFocus = iter;
	  }
	  drawRequired = True;
	}
      }
      else			/* update flags */
      {
	/* mark this task */
	MarkTask(iter);
	if (iter->dwFlags != flags) /* have the flags changed? */
	{
	  iter->dwFlags = flags;
	  drawRequired = True;
	}
      }
    }
    XFree(pWin);
    /* now remove all unmarked tasks */
    /* check to see if the current focus is not marked, if not we'll 
       remove it 
    */
    if (currentFocus && !(currentFocus->dwFlags & TF_VALID))
    {
      currentFocus = NULL;
    }
    
    if (RemoveUnmarkedTasks(&taskHead) > 0)
    {
      drawRequired = True;
    }
    /* now let's recalc the numTasks, some hidden ones won't be added */
    RecalcGeometry();
    /* now force a redraw */
    if (drawRequired)
    {
#ifdef DEBUG
      printf("Redrawing in response to NET_CLIENT_LIST\n");
#endif
      Draw();
    }
  }
}

static void OnDestroy(Window window)
{
#ifdef DEBUG
  printf("In OnDestroy 0x%08X\n",(unsigned int)window);
#endif
  /* remove currentFocus is we have to */
  if (currentFocus && (currentFocus->wnd == window))
  {
    currentFocus = NULL;
  }
  /* try to remove the item */
  RemoveTask(&taskHead,window);
  /* now recalculate */
  RecalcGeometry();
  /* now redraw */
  Draw();
}

static void OnPropertyNotify(Window window,Atom atom)
{
  FINDTASK ft;
  LPTASK pTask;
  int iconified;
  
  if (window == root)
  {
    if (atom == atom__NET_CURRENT_DESKTOP)
    {
      /* free all tasks, totally rebuild */
#ifdef DEBUG
      printf("NET_CURRENT_DESKTOP has changed, recomputing tasks\n");
#endif
      gCurrentDesktop = GetCurrentDesktop();
      /* remove the focus */
      currentFocus = NULL;
      /* continue now */
      RemoveAllTasks(&taskHead);
      BuildClientList(1);
      return;
    }
    else if (atom == atom__NET_CLIENT_LIST)
    {
      BuildClientList(0);
      return;
    }
  }
  /* try to find the task here */
  ft.dwFlags = FTF_WINDOW;
  ft.wnd = window;
  pTask = FindTask(taskHead,&ft);
  if (!pTask)
  {
    return;
  }
#ifdef DEBUG
  printf("PropertyNotify for window 0x%08X\n",(unsigned int)window);
#endif
  if (atom == XA_WM_NAME)
  {
    UpdateName(pTask);
    Redraw(pTask);
  }
  else if (atom == atom__WM_STATE)
  {
    iconified = (pTask->dwFlags & TF_ICONIFIED) ? 1 : 0; 
    if (IsIconified(pTask->wnd) != iconified)
    {
      pTask->dwFlags = (pTask->dwFlags & ~TF_ICONIFIED) | (!iconified ? TF_ICONIFIED : 0);
      Redraw(pTask);
    }
  }
  else if (atom == XA_WM_HINTS)	/* set after mapping, update icon */
  {
    LoadIcon(pTask);
    Redraw(pTask);
  }
}

static void OnFocus(Window window)
{
  LPTASK iter;
  FINDTASK ft;
  /* int pid; */
#ifdef DEBUG
  printf("Window 0x%08x received focus\n",(unsigned int)window);
#endif
  
  if (currentFocus)
  {
    currentFocus->dwFlags &= ~TF_FOCUS;
    Redraw(currentFocus);
  }

  /*pid = GetWindowPid(window);
  if (pid > 0)
  {
    printf("Window 0x%08X has pid %i\n",window,pid);
    if (setpriority(PRIO_PROCESS,pid,-20) == 0)
    {
      printf("Successfully reniced pid %i to -20\n",pid);
    }
    else
    {
      printf("Failed to renice pid %i to -20\n",pid);
    }
  }*/

  /* now find the task */
  ft.dwFlags = FTF_WINDOW;
  ft.wnd = window;
  iter = FindTask(taskHead,&ft);
  if (iter)
  {
    iter->dwFlags |= TF_FOCUS;
    currentFocus = iter;
    Redraw(currentFocus);
  }
}

static int HitTest(XEvent *ev)
{
  int item;
  int slack;
  /* only allow clicks on the button */
  if (ev->xbutton.y < OFFSET || ev->xbutton.y > (WND_HEIGHT - OFFSET))
  {
    return -1;
  }
  
  item = (ev->xbutton.x - startButtonAreaWidth) / buttonWidth;
  slack = (ev->xbutton.x - startButtonAreaWidth) % buttonWidth;
#ifdef DEBUG
  printf("Slack is %i\n",slack);
#endif
  if (item >= numTasks || (slack < OFFSET) || (slack > (buttonWidth - OFFSET)))
  {
    return -1;
  }
  return item;
}

static void OnMouseDown(XEvent *ev)
{
  int item;
  int x;
  int y;
  LPTASK pTask;
  PROGRAMMENUOPTIONS pmop;

  item = HitTest(ev);
  if (item >= 0)
  {
#ifdef DEBUG
    printf("Clicked item %i\n",item);
#endif
    /* now select/deselect it */
    /* get the task here */
    pTask = GetTask(taskHead,item);
    if (pTask)
    {
      if (ev->xbutton.button == 3) /* shading */
      {
	if (!(pTask->dwFlags & TF_ICONIFIED))
	{
	  ToggleShade(pTask->wnd);
	}
      }
      else
      {
	if (pTask->dwFlags & TF_ICONIFIED) /* if minimized, maximize! */
	{
	  Maximize(pTask->wnd);
	  XSync(pDisp,False);
	  /* I call this twice.  Seems that my window manager sometimes
	   has trouble with a maximize + set focus all at once.  Using
	  two calls fixes it */
	  XSync(pDisp,False); 
	}
	if (!(pTask->dwFlags & TF_FOCUS)) /* now set focus */
	{
#ifdef DEBUG
	  printf("Calling SetFocus\n");
#endif
	  SetFocus(pTask->wnd);
	}
      }
    }
  }
  else				/* didn't hit, check if hit the start menu */
  {
    if (useMenu && ev->xbutton.button == 1 && ev->xbutton.x < startButtonAreaWidth)
    {
      if (!pMenu && pConfigFile)
      {
	pmop.pList = pConfigFile;
	pmop.parent = mainWindow;
	pmop.useIcons = useMenuIcons;
	pmop.ms = menuStyle;
	pMenu = CreateProgramMenu(&pmop);
      }

      if (pMenu)
      {
	ClientToScreen(mainWindow,OFFSET,0,&x,&y);
	if (onTop)
	{
	  y += WND_HEIGHT;
	}

	ShowMenu(pMenu,onTop ? MENU_SHOW_TOP_LEFT : MENU_SHOW_BOTTOM_LEFT,x,y);
	/* redraw the start menu */
	DrawStartButton();
      }
    }
  }
}

static void LibraryTimeout()
{
  LPLIBENTRY pIter;

  for (pIter = (LPLIBENTRY)pLibList->pHead;pIter;pIter = (LPLIBENTRY)pIter->node.pNext)
  {
    if (pIter->pModule->Timeout)
    {
      pIter->pModule->Timeout(pIter->pModule);
    }
  }
}

static void DrawModules()
{
  LPLIBENTRY pIter;
  int i;
  for (pIter = (LPLIBENTRY)pLibList->pHead;pIter;pIter = (LPLIBENTRY)pIter->node.pNext)
  {
    for (i=0;i<pIter->pModule->numPanels;++i)
    {
      if (pIter->pModule->panels[i].wnd)
      {
	/* dispatch and return */
	XClearArea(pDisp,pIter->pModule->panels[i].wnd,0,0,0,0,True);
      }
    }
  }
}

static void DispatchToLibraries(XEvent *event)
{
  /* search through the window list */
  LPLIBENTRY pIter;
  struct itimerval tval;
  int i;
  for (pIter = (LPLIBENTRY)pLibList->pHead;pIter;pIter = (LPLIBENTRY)pIter->node.pNext)
  {
    for (i=0;i<pIter->pModule->numPanels;++i)
    {
      if (pIter->pModule->panels[i].wnd == event->xany.window)
      {
	/* check for custom things like tooltips */
	if (useTooltips)
	{
	  if (event->type == EnterNotify)
	  {
	    if (!pTooltip)
	    {
	      pTooltipModule = pIter->pModule;
	      /* start the timer to fire one shot */
	      signal(SIGALRM,AlarmHandler);
	      memset(&tval,0,sizeof(tval));
	      tval.it_value.tv_usec = 1e6/2;
	      setitimer(ITIMER_REAL,&tval,NULL);
	    }
	  }
	  else if (event->type == LeaveNotify)
	  {
	    /* stop the timer */
	    memset(&tval,0,sizeof(tval));
	    setitimer(ITIMER_REAL,&tval,NULL);
	    
	    pTooltipModule = NULL;
	    
	    if (pTooltip)
	    {
	      DestroyTooltip(pTooltip);
	      pTooltip = NULL;
	    }
	  }
	}
	/* dispatch and return */
	pIter->pModule->EventHandler(pIter->pModule,event);
	return;
      }
    }
  }
}

Window CreatePanel()
{
  XSetWindowAttributes att;
  unsigned long strut[4];
  MWMHints mwm;
  XSizeHints sizeHints;
  XWMHints wmHints;
  XGCValues gcv;
  unsigned int mask;
  int dummy;
  char *exampleString;
  FONTRECT fontRect;

#ifdef HAVE_XINERAMA
  /* 
     int major,minor;
  XineramaQueryVersion(pDisp,&major,&minor);
  XineramaIsActive(pDisp);
  */
#endif

  att.background_pixel = PALETTE_LIGHT_GRAY;
  att.event_mask = ButtonPressMask | ButtonReleaseMask | ExposureMask;
  mainWindow = XCreateWindow(pDisp,root,0,gScreenHeight - WND_HEIGHT,gScreenWidth,WND_HEIGHT,0,CopyFromParent,InputOutput,CopyFromParent,CWEventMask | CWBackPixel,&att);
  
  /* live on all desktops */
  SetProperty(mainWindow,atom__NET_WM_DESKTOP,XA_CARDINAL,0xFFFFFFFF);
  SetProperty(mainWindow,atom__NET_WM_WINDOW_TYPE,XA_ATOM,atom__NET_WM_WINDOW_TYPE_DOCK);
  /* skip this window in the window list */
  SetProperty(mainWindow,atom__WIN_HINTS,XA_CARDINAL,
	      WIN_HINTS_SKIP_FOCUS |
	      WIN_HINTS_SKIP_WINLIST |
	      WIN_HINTS_SKIP_TASKBAR |
	      WIN_HINTS_DO_NOT_COVER);
	      
  /* borderless motif */
  memset(&mwm,0,sizeof(mwm));
  mwm.flags = MWM_HINTS_DECORATIONS;
  XChangeProperty(pDisp,mainWindow,atom__MOTIF_WM_HINTS,atom__MOTIF_WM_HINTS,32,PropModeReplace,(unsigned char*)&mwm,sizeof(MWMHints)/4);
  /* make sure WM obeys our window position */
  sizeHints.flags = PPosition;
  XChangeProperty(pDisp,mainWindow,XA_WM_NORMAL_HINTS,XA_WM_SIZE_HINTS,32,PropModeReplace,(unsigned char*)&sizeHints,sizeof(XSizeHints)/4);
  /* make unfocusable */
  wmHints.flags = InputHint;
  wmHints.input = False;
  XChangeProperty(pDisp,mainWindow,XA_WM_HINTS,XA_WM_HINTS,32,PropModeReplace,(unsigned char*)&wmHints,sizeof(XWMHints)/4);
  
  /* load the fonts */
#ifdef HAVE_XFT
  if (useXft)
  {
    pFont = CreateXFTFont(pDisp,gScreen,mainWindow,fontName,0);
  }
  else
  {
    pFont = CreateXFont(pDisp,gScreen,mainWindow,fontName,0);
  }
#else
  pFont = CreateXFont(pDisp,gScreen,mainWindow,fontName,0);
#endif
  
  /* determine WND_HEIGHT now */
  exampleString = "THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG";
  pFont->GetExtents(pFont,exampleString,strlen(exampleString),&fontRect);
  
  WND_HEIGHT = max(ICON_HEIGHT,max(pFont->ascent,fontRect.height)) + 5*OFFSET;
  /* now resetup WND_HEIGHT */
  XMoveResizeWindow(pDisp,mainWindow,0,onTop ? 0 : gScreenHeight - WND_HEIGHT,gScreenWidth,WND_HEIGHT);
  /* reserve "WINHEIGHT" pixels at the bottom of the screen */
  strut[0] = 0;
  strut[1] = 0;
  strut[2] = onTop ? WND_HEIGHT : 0;
  strut[3] = onTop ? 0 : WND_HEIGHT;
  XChangeProperty (pDisp, mainWindow, atom__NET_WM_STRUT, XA_CARDINAL, 32,
		   PropModeReplace, (unsigned char *) strut, 4);

  pFont->GetCenter(pFont,exampleString,strlen(exampleString),0,WND_HEIGHT,&dummy,&textY);
		   
  /* finish the rest of initialization */
  gcv.graphics_exposures = True;
  gcv.font = pFont->GetID(pFont);
  if (gcv.font >= 0)
  {
    mask = GCFont | GCGraphicsExposures;
  }
  else
  {
    mask = GCGraphicsExposures;
  }
  
  gGc = XCreateGC(pDisp,mainWindow,mask,&gcv);
  /* create pixmaps */
  LoadPixmap(icon_xpm,&defaultPixmap);
  LoadPixmap(start_xpm,&startMenuIcon);
  /* determine length of start menu area */
  if (useMenu)
  {
    startButtonAreaWidth = startMenuIcon.width + pFont->GetWidth(pFont,"Start",5) + 5 * OFFSET;
  }
  /* display at the end */
  XMapWindow(pDisp,mainWindow); 
  return mainWindow;
}

static void LoadModules()
{
  int current = gScreenWidth - OFFSET;
  int i;
  int numWnds=0;
  Window root_ret,tmp;
  int x,y;
  unsigned int ww,hh,borderWidth,depth;
  LPLIBENTRY pIter;
  LIBRARYINIT linit;
  LPCONFIGITEM pItem;
  LPCONFIGITEM pOptions;
  LPCONFIGITEM pPath;
  /* DIR *pDir;
  struct dirent *pDirent;
  char *pExt;
  */

  /* create list now */
  pLibList = LibListCreate();

  linit.pDisplay = pDisp;
  linit.screen = gScreen;
  linit.root = root;
  linit.panelWindow = mainWindow;
  linit.height = SUBWND_HEIGHT;
  linit.pFont = pFont;
  linit.backgroundPixel = PALETTE_LIGHT_GRAY; 
  linit.pModule= NULL;
  linit.pixmapFuncs.LoadPixmap = LoadPixmap;
  linit.pixmapFuncs.LoadPixmapFromFile = LoadPixmapFromFile;
  linit.pixmapFuncs.FreePixmap = FreePixmap;
  linit.pixmapFuncs.UpdatePixmapAttributes = UpdatePixmapAttributes;
  linit.pixmapFuncs.RenderPixmap = RenderPixmap;
  linit.pixmapFuncs.ScalePixmap = ScalePixmap;
  linit.baseInputMask = (useTooltips ? (EnterWindowMask | LeaveWindowMask) : 0);
  /* get modules from config file */
  if (pConfigFile && 
      (pItem = FindConfigItemByName(pConfigFile,"plugins")) && 
      pItem->pSublist)
  {
    for (pItem = (LPCONFIGITEM)pItem->pSublist->list.pHead;pItem;pItem = (LPCONFIGITEM)pItem->node.pNext)
    {
      if (pItem->pSublist)
      {
	pPath = FindConfigItemByName(pItem->pSublist,"path");
	pOptions = FindConfigItemByName(pItem->pSublist,"options");
	if (pPath)
	{
	  linit.pOptions = pOptions;
	  /* now add */
	  LibListAdd(pLibList,&linit,pPath->value);
	}
      }
    }
  }
  /* now I will place them accordingly */
  for (pIter = (LPLIBENTRY)pLibList->pHead;pIter;pIter = (LPLIBENTRY)pIter->node.pNext)
  {
    /* iterate through all windows */
    for (i=0;i<pIter->pModule->numPanels;++i)
    {
      tmp = pIter->pModule->panels[i].wnd;
      /* determainWindow the size of it */
      XGetGeometry(pDisp,tmp,&root_ret,&x,&y,&ww,&hh,&borderWidth,&depth);
      /* now move the current pointer */
      current -= ww;
      
      XMoveWindow(pDisp,tmp,current,(WND_HEIGHT - SUBWND_HEIGHT)/2+y);
#ifdef DEBUG
      printf("Found a window, moving it to %ix%i (%ix%i)\n",current,y,ww,hh);
#endif
      /* now show it */
      XRaiseWindow(pDisp,tmp);
      /* increment count */
      ++numWnds;
    }
  }
  
  XMapSubwindows(pDisp,mainWindow);
  
  /* modules were loaded */
  if (numWnds > 0)
  {
    xModuleStart = current - 1;
    xModuleEnd   = gScreenWidth - OFFSET;
  }
  else				/* no modules loaded */
  {
    xModuleStart = xModuleEnd = gScreenWidth;
  }
}

/* forks and kills the parent, child will continue with /dev/null as input
   and output
*/
static void mydaemon()
{
  switch (fork())
  {
  case -1:
    fprintf(stderr,"Failed to fork, exiting\n");
    Handler(SIGTERM);
    break;
  case 0:
    /* i am the child, redirect my output, and continue */
    close(0);
    close(1);
    close(2);
    
    open("/dev/null",O_RDWR); /* this will return 0 */
    dup(0);			/* this will return 1 */
    dup(0);			/* this will return 2 */
    break;
  default:
    /* i am the parent, i will quit */    
    exit(0);
  }
}

#ifdef __linux__
#define SELECTWRAPPER select
#else
static void subtractTimeval(struct timeval *left,struct timeval *right,
			    struct timeval *result)
{
  /* subtract the seconds */
  result->tv_sec = left->tv_sec - right->tv_sec;
  result->tv_usec = left->tv_usec - right->tv_usec;
  /* modify the seconds if the microseconds are below 0 */
  while (result->tv_usec < 0)
  {
    result->tv_sec--;
    result->tv_usec += 1000000;
  }
  /* now do some sanity checks */
  if (result->tv_sec < 0)
  {
    result->tv_sec = 0;
    result->tv_usec = 0;
  }
}

static int SELECTWRAPPER(int n,fd_set *read,fd_set *write,fd_set *except,
		  struct timeval *timeout)
{
  int ret;
  struct timeval start,end,elapsed,orig;
  struct timezone tz;
  /* save the original wait time, select might modify it in some unknown way */
  memcpy(&orig,timeout,sizeof(struct timeval));
  /* get start time */
  gettimeofday(&start,&tz);
  /* make the call */
  ret = select(n,read,write,except,timeout);
  /* get the ending time */
  gettimeofday(&end,&tz);
  /* now compute the new timeout */
  subtractTimeval(&end,&start,&elapsed);
  subtractTimeval(&orig,&elapsed,timeout);
  return ret;
}
#endif

int main(int argc,char **argv,char **env)
{
  Window mainWindow;
  XEvent ev;
  int fd;
  fd_set readSet;
  struct timeval tval;
  int destroyed;

  globalArgc = argc;
  globalArgv = argv;
  globalEnv  = env;

  lt_dlinit();

  InstallSignals();
  
  XSetErrorHandler(ErrorHandler);
  XSetIOErrorHandler(IOErrorHandler);

  pDisp = XOpenDisplay(NULL);
  if (!pDisp)
  {
    PrintDisplayError();
    return 1;
  }

  LoadAtoms(pDisp);

  gScreen = DefaultScreen(pDisp);
  root = RootWindow(pDisp,gScreen);
  gScreenWidth = DisplayWidth(pDisp,gScreen);
  gScreenHeight = DisplayHeight(pDisp,gScreen);
  gScreenDepth  = DefaultDepth(pDisp,gScreen);
  gCurrentDesktop = GetCurrentDesktop(pDisp);

  if (!LoadSettings(argc,argv))
  {
    XCloseDisplay(pDisp);
    return 67;
  }
  
  /* fork is requested */
  if (useDaemon)
  {
    mydaemon();
  }
  
#ifdef DEBUG
  printf("Opened display, settings are %ix%i:%i\n",gScreenWidth,gScreenHeight,gScreenDepth);
#endif
  
  InitPixmap();
  /* now create a window */
  LoadPalette();
  mainWindow = CreatePanel();
 
  LoadModules();
  
  /* Hook up to root window */
  XSelectInput(pDisp,root,PropertyChangeMask);
  BuildClientList(1);
 
  /* Draw Modules for the first time  */
  DrawModules();

  XFlush(pDisp);

  fd = ConnectionNumber(pDisp);
  tval.tv_sec = 1;		/* setup the timer */
  tval.tv_usec = 0;
  while (1)
  {
    FD_ZERO(&readSet);
    FD_SET(fd,&readSet);
    /* use a wrapper which emulates Linux's version that updates the tval
       to the time remaining */
    if (SELECTWRAPPER(fd+1,&readSet,0,0,&tval))	
    {
      while (XPending(pDisp))
      {
	XNextEvent(pDisp,&ev);
	/* process menu */
	if (pMenu && ProcessMenu(pMenu,&ev,&destroyed))
	{
	  if (destroyed) 
	  {
	    /* redraw if the menu closes */
	    DrawStartButton();
	  }
	}
	else if (pTooltip && ProcessTooltip(pTooltip,&ev))
	{
	  /* continue on, skip processing the modules */
	}
	else /* process regular things */
	{
	  switch (ev.type)
	  {
	  case DestroyNotify:
	    OnDestroy(ev.xdestroywindow.window);
	    break;
	  case PropertyNotify:
	    OnPropertyNotify(ev.xproperty.window,ev.xproperty.atom);
	    break;
	  case FocusIn:
	    OnFocus(ev.xfocus.window);
	    break;
	  case Expose:
	    if (ev.xexpose.window == mainWindow && ev.xexpose.count == 0)
	    {
	      Draw();
	      continue;
	    }
	    break;
	  case ButtonPress:
	    if (ev.xbutton.window == mainWindow)
	    {
	      OnMouseDown(&ev);
	      continue;
	    }
	    break;
	  }
	  /* dispatch to libraries */
	  DispatchToLibraries(&ev);
	}
      }
    }
    else
    {
      LibraryTimeout();
      /* reset the timer */
      tval.tv_sec = 1;
      tval.tv_usec = 0;
    }
  }
  /* this will never be reached */
  return 0;
}
